Imports System.IO

Public Class CStatSum
  Private Const EPSILON = 1.0E-20
  Private Const DIGIT_MARKERS = "-+0123456789."

  Private m_bZeroX As Boolean
  Private m_bZeroY As Boolean
  Private m_bNegX As Boolean
  Private m_bNegY As Boolean

  Private m_fSum As Double
  Private m_fSumX As Double
  Private m_fSumX2 As Double
  Private m_fSumY As Double
  Private m_fSumY2 As Double
  Private m_fSumXY As Double
  Private m_fMeanX As Double
  Private m_fMeanY As Double
  Private m_fSdevX As Double
  Private m_fSdevY As Double
  Private m_fSlope As Double
  Private m_fIntercept As Double
  Private m_fR2 As Double
  Private m_sTX As String
  Private m_sTY As String
  Private m_sWt As String

  Private m_fX() As Double
  Private m_fY() As Double
  Private m_fWt() As Double
  Private m_fShiftX As Double
  Private m_fShiftY As Double
  Private m_fScaleX As Double
  Private m_fScaleY As Double
  Private m_nDataCount As Integer


  Public Sub InitSums()
    m_fSum = 0
    m_fSumX = 0
    m_fSumX2 = 0
    m_fSumY = 0
    m_fSumY2 = 0
    m_fSumXY = 0
  End Sub

  Public Function GetData(ByVal sDataFilename As String, ByRef nDataCount As Integer, _
     Optional ByVal ShiftX As Double = 0, Optional ByVal ShiftY As Double = 0, _
     Optional ByVal ScaleX As Double = 0, Optional ByVal ScaleY As Double = 1) As Boolean

    Dim sLine, sLines() As String
    Dim sData() As String
    Dim I, J, K, N As Integer
    Dim bRes As Boolean = True

    Try
      sLines = File.ReadAllLines(sDataFilename)

      nDataCount = sLines.GetUpperBound(0)

      ' Dimension arrays for maximum capacity
      ReDim m_fX(nDataCount)
      ReDim m_fY(nDataCount)
      ReDim m_fWt(nDataCount)

      J = 0
      m_nDataCount = 0
      Do Until J = sLines.Length
        sLine = sLines(J).Trim()
        ' is line not empty?
        If sLine.Length > 0 Then
          ' is it NOT a comment?
          If DIGIT_MARKERS.IndexOf(sLine.Substring(0, 1)) >= 0 Then
            sData = sLine.Split(",")
            N = sData.GetUpperBound(0)
            m_fX(m_nDataCount) = Double.Parse(sData(0))
            m_fY(m_nDataCount) = Double.Parse(sData(1))
            If N < 2 Then
              m_fWt(m_nDataCount) = 1
            Else
              m_fWt(m_nDataCount) = Double.Parse(sData(2))
            End If
            m_nDataCount += 1
          End If
        End If
        J += 1
      Loop


      ' adjust arrays to actual number of data
      ReDim Preserve m_fX(m_nDataCount)
      ReDim Preserve m_fY(m_nDataCount)
      ReDim Preserve m_fWt(m_nDataCount)
      nDataCount = m_nDataCount

      For I = 0 To m_nDataCount - 1
        m_fX(I) = ScaleX * (m_fX(I) - ShiftX)
        m_fY(I) = ScaleY * (m_fY(I) - ShiftY)
      Next I

    Catch ex As Exception
      bRes = False
    End Try

    Return bRes

  End Function

  Public Sub New()
    InitSums()
  End Sub

  Private Sub Add(ByVal X As Double, ByVal Y As Double, Optional ByVal Wt As Double = 1)
    m_fSum = m_fSum + Wt
    m_fSumX = m_fSumX + X * Wt
    m_fSumX2 = m_fSumX2 + X * X * Wt
    m_fSumY = m_fSumY + Y * Wt
    m_fSumY2 = m_fSumY2 + Y * Y * Wt
    m_fSumXY = m_fSumXY + X * Y * Wt
  End Sub

  Public Sub FindBestFit(ByRef objRes As CResults)

    Dim I As Integer
    Dim ITX As FitType
    Dim ITY As FitType
    Dim bOK As Boolean
    Dim fXt, fYt As Double
    Dim objErrs As New CErrors

    If m_nDataCount < 3 Then Exit Sub

    objRes.Clear()

    Try
      m_bZeroX = False
      m_bZeroY = False
      m_bNegX = False
      m_bNegY = False

      'objRes.Initialize

      For I = 0 To m_nDataCount - 1
        If m_fX(I) < 0 Then m_bNegX = True
        If m_fY(I) < 0 Then m_bNegY = True
        If Math.Abs(m_fX(I)) < EPSILON Then m_bZeroX = True
        If Math.Abs(m_fY(I)) < EPSILON Then m_bZeroY = True
      Next I

      m_sWt = ""

      For ITY = TypeModule.FitType.eLinear To TypeModule.FitType.eLn
        ' validate transformations
        If m_bZeroY And m_bNegY Then
          bOK = CanHandleZeroAndNegative(ITY)
        ElseIf m_bZeroY Then
          bOK = CanHandleZero(ITY)
        ElseIf m_bNegY Then
          bOK = CanHandleNegative(ITY)
        Else
          bOK = True
        End If

        ' Can proceed?
        If bOK Then

          For ITX = TypeModule.FitType.eLinear To TypeModule.FitType.eLn

            ' validate transformations
            If m_bZeroX And m_bNegX Then
              bOK = CanHandleZeroAndNegative(ITX)
            ElseIf m_bZeroX Then
              bOK = CanHandleZero(ITX)
            ElseIf m_bNegX Then
              bOK = CanHandleNegative(ITX)
            Else
              bOK = True
            End If

            ' Can proceed?
            If bOK Then

              InitSums()

              ' store transformation data
              m_sTX = SayTransform(ITX)
              m_sTY = SayTransform(ITY, "Y")

              For I = 0 To m_nDataCount - 1

                Select Case ITX
                  Case TypeModule.FitType.eLinear
                    fXt = m_fX(I)
                  Case TypeModule.FitType.eSquare
                    fXt = m_fX(I) ^ 2
                  Case TypeModule.FitType.eCube
                    fXt = m_fX(I) ^ 3
                  Case TypeModule.FitType.eCubeRoot
                    fXt = m_fX(I) ^ (1 / 3)
                  Case TypeModule.FitType.eRecip
                    fXt = 1 / m_fX(I)
                  Case TypeModule.FitType.eRecipCubeRoot
                    fXt = 1 / m_fX(I) ^ (1 / 3)
                  Case TypeModule.FitType.eRecipSquare
                    fXt = 1 / m_fX(I) ^ 2
                  Case TypeModule.FitType.eRecipCube
                    fXt = 1 / m_fX(I) ^ 3
                  Case TypeModule.FitType.eSqrt
                    fXt = Math.Sqrt(m_fX(I))
                  Case TypeModule.FitType.eRecipSqrt
                    fXt = 1 / Math.Sqrt(m_fX(I))
                  Case TypeModule.FitType.eLn
                    fXt = Math.Log(m_fX(I))
                End Select


                Select Case ITY
                  Case TypeModule.FitType.eLinear
                    fYt = m_fY(I)
                  Case TypeModule.FitType.eSquare
                    fYt = m_fY(I) ^ 2
                  Case TypeModule.FitType.eCube
                    fYt = m_fY(I) ^ 3
                  Case TypeModule.FitType.eCubeRoot
                    fYt = m_fY(I) ^ (1 / 3)
                  Case TypeModule.FitType.eRecip
                    fYt = 1 / m_fY(I)
                  Case TypeModule.FitType.eRecipCubeRoot
                    fYt = 1 / m_fY(I) ^ (1 / 3)
                  Case TypeModule.FitType.eRecipSquare
                    fYt = 1 / m_fY(I) ^ 2
                  Case TypeModule.FitType.eRecipCube
                    fYt = 1 / m_fY(I) ^ 3
                  Case TypeModule.FitType.eSqrt
                    fYt = Math.Sqrt(m_fY(I))
                  Case TypeModule.FitType.eRecipSqrt
                    fYt = 1 / Math.Sqrt(m_fY(I))
                  Case TypeModule.FitType.eLn
                    fYt = Math.Log(m_fY(I))
                End Select

                ' add transformed data to statistical summations
                Add(fXt, fYt, m_fWt(I))

              Next I

              objErrs.Clear()

              ' calculate regression statistics and store in
              ' object accessed by m_objRes
              CalcLR(objErrs)

              If objErrs.GetCount > 1 Then
                objRes.Add(m_sTY & " = A + B * " & m_sTX, m_fR2, m_fSlope, m_fIntercept, objErrs.GetErrText(0))
              Else
                objRes.Add(m_sTY & " = A + B * " & m_sTX, m_fR2, m_fSlope, m_fIntercept, "")
              End If

            End If

          Next ITX

        End If
      Next ITY

    Catch ex As Exception
      objErrs.Add(ex.Message)
    End Try
  End Sub



  Private Function SayTransform(ByVal eVal As FitType, Optional ByVal sVar As String = "X") As String
    Select Case eVal
      Case TypeModule.FitType.eLinear
        Return sVar
      Case TypeModule.FitType.eSquare
        Return sVar & "^2"
      Case TypeModule.FitType.eCube
        Return sVar & "^3"
      Case TypeModule.FitType.eCubeRoot
        Return sVar & "^1/3"
      Case TypeModule.FitType.eRecip
        Return "1/" & sVar
      Case TypeModule.FitType.eRecipCubeRoot
        Return "1/" & sVar & "^1/3"
      Case TypeModule.FitType.eRecipSquare
        Return "1/" & sVar & "^2"
      Case TypeModule.FitType.eRecipCube
        Return "1/" & sVar & "^3"
      Case TypeModule.FitType.eSqrt
        Return sVar & "^1/2"
      Case TypeModule.FitType.eRecipSqrt
        Return "1/" & sVar & "^1/2"
      Case TypeModule.FitType.eLn
        Return "Ln(" & sVar & ")"
    End Select
  End Function


  Private Function CanHandleZero(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TypeModule.FitType.eLinear, TypeModule.FitType.eSquare, TypeModule.FitType.eCube, _
                    TypeModule.FitType.eCubeRoot, TypeModule.FitType.eSqrt
        CanHandleZero = True
      Case Else
        CanHandleZero = False
    End Select
  End Function

  Private Function CanHandleZeroAndNegative(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TypeModule.FitType.eLinear, TypeModule.FitType.eSquare, TypeModule.FitType.eCube
        CanHandleZeroAndNegative = True
      Case Else
        CanHandleZeroAndNegative = False
    End Select
  End Function

  Private Function CanHandleNegative(ByVal eVal As FitType) As Boolean

    Select Case eVal
      Case TypeModule.FitType.eLinear, TypeModule.FitType.eSquare, TypeModule.FitType.eCube, _
            TypeModule.FitType.eRecip, TypeModule.FitType.eRecipSquare, TypeModule.FitType.eRecipCube
        CanHandleNegative = True
      Case Else
        CanHandleNegative = False
    End Select
  End Function


  Private Sub CalcLR(ByRef objErrs As CErrors)

    If m_fSum < 2 Then Exit Sub
    ' caluclate regression
    Try
      m_fMeanX = m_fSumX / m_fSum
      m_fMeanY = m_fSumY / m_fSum
      m_fSdevX = Math.Sqrt((m_fSumX2 - m_fSumX ^ 2 / m_fSum) / (m_fSum - 1))
      m_fSdevY = Math.Sqrt((m_fSumY2 - m_fSumY ^ 2 / m_fSum) / (m_fSum - 1))
      m_fSlope = (m_fSum * m_fSumXY - m_fSumX * m_fSumY) / _
                 (m_fSum * m_fSumX2 - m_fSumX ^ 2)
      m_fIntercept = m_fMeanY - m_fSlope * m_fMeanX
      m_fR2 = ((m_fSum * m_fSumXY - m_fSumX * m_fSumY) / _
               (m_fSum * (m_fSum - 1) * m_fSdevX * m_fSdevY)) ^ 2

    Catch ex As Exception
      objErrs.Add("For model" & m_sTY & " = A + B * " & m_sTX & vbCrLf & ex.Message)
    End Try

  End Sub
End Class
